"""
Machine deployer

Deploy a GCE machine with the informed params
"""

from googleapiclient import discovery
from googleapiclient.errors import HttpError

from gcpinfra.gcp.conf import GCPConf
from gcpinfra.gce.conf import GCEMachine

# pylint: disable=invalid-name

VALID_STATUS = ['PROVISIONING', 'STAGING', 'RUNNING', 'STOPPING',
                'SUSPENDING', 'SUSPENDED', 'REPAIRING', 'TERMINATED']
DELETED = 'DELETED'

class GCEMachineDeployer:
    """
    Represents a machine deployer.
    """

    def __init__(self, name, zone, machine, docker_image, env=None,
                 restart_policy='Always'):
        """Constructor."""

        self.conf = GCPConf()

        self.machine = machine
        self.name = name
        self.zone = zone
        self.docker_image = docker_image
        self.restart_policy = restart_policy
        self.region = self.__get_region()  # only after zone declaration
        self.env = env
        self.status_rep = None
        self.status = None
        self.was_instantiated = False

        if self.restart_policy not in ['OnFailure', 'Never', 'Always']:
            raise ValueError("Invalid 'restart_policy' value")

        if not isinstance(self.machine, GCEMachine):
            raise ValueError("Expected a machine 'GCEMachine' got '{}'".format(
                type(self.machine)))

        if self.env is not None and not isinstance(self.env, dict):
            raise ValueError("Expected a 'dict' or 'NoneType' got '{}'".format(
                type(self.env)))
        elif self.env:
            pitem = "        - name: {}\n          value: '{}'\n"
            self.env = '      env:\n{}'.format(
                ''.join([pitem.format(*l) for l in self.env.items()]))
        else:
            self.env = ''

        while True:
            try:
                self.compute = discovery.build('compute', 'v1')
            except HttpError as e:
                if int(e.resp['status']) == 500:
                    continue
            break

    def __get_region(self):
        """Returns the region."""

        return '-'.join(self.zone.split('-')[:-1])

    def __metadata_items(self):
        """Retorna os itens de metadata."""

        val = 'spec:\n  containers:\n    - name: {}\n      ' \
              'image: {}\n{}      stdin: false\n      tty: ' \
              'false\n  restartPolicy: {}\n\n# This container' \
              ' declaration format is not public API and may ' \
              'change without notice. Please\n# use gcloud ' \
              'command-line tool or Google Cloud Console to ' \
              'run Containers on Google Compute Engine.'.format(
                 self.name, self.docker_image, self.env,
                 self.restart_policy)
        return val

    def instantiate(self):
        """Deploy a machine."""

        try:
            self.compute.instances().insert(
                project=self.conf.project_id, zone=self.zone,
                body=self.mount_representation()).execute()

            self.get()
            if self.status in VALID_STATUS:
                self.was_instantiated = True
        except HttpError as e:
            if int(e.resp['status']) == 500:
                self.instantiate()  # tenta novamente em caso de erro 500

    def get(self):
        """Get updated representation."""

        try:
            self.status_rep = self.compute.instances().get(
                project=self.conf.project_id, zone=self.zone,
                instance=self.name).execute()
            self.status = self.status_rep['status']
        except HttpError as e:
            self.status_rep = None
            # check if this was instantiated at some point and the http return
            # code is 404. If so, this instance was killed/deleted
            if self.was_instantiated and int(e.resp['status']) == 404:
                self.status = 'DELETED'
            elif int(e.resp['status']) == 500:
                self.get()  # tenta novamente em caso de erro 500

    def mount_representation(self):
        """Mount this object representation."""

        _rep = {
            'kind': 'compute#instance',
            'name': self.name,
            'zone': 'projects/{}/zones/{}'.format(self.conf.project_id, self.zone),
            'machineType': 'projects/{}/zones/{}/machineTypes/{}'.format(
                self.conf.project_id, self.zone, self.machine.machine_type_uri),
            'displayDevice': {'enableDisplay': False},
            'metadata': {
                'kind': 'compute#metadata',
                'items': [
                    {
                        'key': 'gce-container-declaration',
                        'value': self.__metadata_items()
                    },
                    {
                        'key': 'google-logging-enabled',
                        'value': 'true'
                    }
                ]
            },
            'tags': {'items': []},
            'disks': [
                {
                    'kind': 'compute#attacheDisk',
                    'type': 'PERSISTENT',
                    'boot': True,
                    'mode': 'READ_WRITE',
                    'autoDelete': True,
                    'deviceName': self.name,
                    'initializeParams': {
                        'sourceImage': ('projects/cos-cloud/global/images/'
                                        'cos-stable-80-12739-91-0'),
                        'diskType': 'projects/{}/zones/{}/diskTypes/{}'.format(
                            self.conf.project_id, self.zone,
                            self.machine.disk_config.boot_disk_type),
                        'diskSizeGb': '{}'.format(
                            self.machine.disk_config.boot_disk_size_gb)
                    },
                    'diskEncryptionKey': {}
                }
            ],
            'canIpForward': False,
            'networkInterfaces': [
                {
                    'kind': 'compute#networkInterface',
                    'subnetwork': 'projects/{}/regions/{}/subnetworks/default'.format(
                        self.conf.project_id, self.region),
                    'accessConfigs': [
                        {
                            'kind': 'compute#accessConfig',
                            'name': 'External NAT',
                            'type': 'ONE_TO_ONE_NAT',
                            'networkTier': 'PREMIUM'
                        }
                    ],
                    'aliasIpRanges': []
                }
            ],
            'description': '',
            'labels': {'container-vm': 'cos-stable-80-12739-91-0'},
            'scheduling': {
                'preemptible': False,
                'onHostMaintenance': 'MIGRATE',
                'automaticRestart': True,
                'nodeAffinities': []
            },
            'deletionProtection': False,
            'reservationAffinity': {'consumeReservationType': 'ANY_RESERVATION'},
            'serviceAccounts': [
                {
                    'email': self.conf.credentials.service_account_email,
                    'scopes': self.conf.scopes
                }
            ],
            'shieldedInstanceConfig': {
                'enableSecureBoot': False,
                'enableVtpm': True,
                'enableIntegrityMonitoring': True
            }
        }
        return _rep



"""
Create clusters to run jobs.

A bunch of configurations are not yet ready to be used.
"""

# pylint: disable=too-few-public-methods, too-many-arguments
# pylint: disable=too-many-instance-attributes

from gcpinfra.gcp.core import GCPCore
from gcpinfra.gce.conf import GCEMachine, GCEDisk

def std_machine(instances=1):
    """Standard machine."""

    return GCEMachine(machine_type='n1-standard-1',
                      instances=instances, accelerators=None,
                      disk=GCEDisk(disk_type='pd-ssd', size=15, num_local_ssds=0))

class GCECluster(GCPCore):
    """
    Cluster to run jobs.

    User must inform the cluster name, configurations and zone to run.
    """

    def __init__(self, cluster_name, master_config=None, worker_config=None,
                 zone_uri=None, image=None, properties=None):
        """Constructor."""

        if not isinstance(cluster_name, str):
            raise ValueError("Param 'cluster_name' must be 'str'")

        self.master_config = master_config
        self.worker_config = worker_config
        if self.master_config is None:
            self.master_config = std_machine()
        if self.worker_config is None:
            self.worker_config = std_machine(instances=2)
        # must always follow master and worker attribution
        self.__validate_machine_configuration()
        self.zone_uri = zone_uri

        if image is None:
            image = '1.3-deb9'

        if properties is None:
            properties = {}

        super().__init__()
        self.add_attr('cluster_name', cluster_name)
        config = {
            'gce_cluster_config': {'zone_uri': zone_uri},
            'master_config': self.master_config.get_rep(),
            'worker_config': self.worker_config.get_rep(),
            'software_config': {
                'image_version': image,
                'properties': properties
            },
            'secondary_worker_config': {
                'num_instances': 0, 'is_preemptible': True
            }
        }
        self.add_attr('config', config)

    def __validate_machine_configuration(self):
        """Verifies if machine is valid."""

        if not isinstance(self.master_config, GCEMachine):
            raise ValueError("Param 'master_config' must be an instance of "
                             "'gcpinfra.gceconf.GCEMachine'")
        if not isinstance(self.worker_config, GCEMachine):
            raise ValueError("Param 'worker_config' must be an instance of "
                             "'gcpinfra.gceconf.GCEMachine'")



"""
Represents a machine in GCE and its configurations.

Those representations will be used during cluster configuration.

Refer to https://cloud.google.com/compute/docs/machine-types for detailed
information
"""

# pylint: disable=too-few-public-methods, invalid-name, too-many-arguments

from gcpinfra.gcp.core import VALID_MACHINES, GCPCore

def valid_machine_type(machine_type):
    """Checks if the informed machine type is valid or not."""

    return machine_type in VALID_MACHINES

class GCEDisk(GCPCore):
    """
    Represents a google compute engine disk.
    """

    def __init__(self, disk_type=None, size=None, num_local_ssds=0):
        """Constructor."""

        super().__init__()
        self.add_attr('boot_disk_type', 'pd-ssd' if disk_type is None else disk_type)
        self.add_attr('boot_disk_size_gb', 15 if size is None else size)
        self.add_attr('num_local_ssds', num_local_ssds)

class GCEMachine(GCPCore):
    """
    Represents a google compute engine machine.
    """

    def __init__(self, machine_type=None, instances=1, disk=None,
                 accelerators=None):
        """Constructor."""

        machine_type = 'n1-standard-1' if machine_type is None else machine_type
        disk = GCEDisk('pd-ssd', 15, 0) if disk is None else disk
        accelerators = [] if accelerators is None else accelerators

        if not isinstance(disk, GCEDisk):
            raise ValueError("Param 'disk' must be an instance of "
                             "'gcpinfra.gcpcore.GCEDisk'")

        if not valid_machine_type(machine_type):
            raise ValueError("Param 'machine_type' ('{}') is not valid. Check "
                             "https://cloud.google.com/compute/docs/machine-types "
                             "for detailed information.".format(machine_type))

        super().__init__()
        self.add_attr('num_instances', instances)
        self.add_attr('machine_type_uri', machine_type)
        self.add_attr('disk_config', disk)
        self.add_attr('accelerators', accelerators)



"""Package install configuration."""

import setuptools

# pylint: disable=invalid-name

version='0.6.1'

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='gcpinfra',
    version=version,
    author='Alex Newman',
    author_email='a.newmanvs@gmail.com',
    description='Google Cloud Platform non-official framework',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/anewmanvs/gcpinfra',
    download_url='https://github.com/anewmanvs/gcpinfra/archive/v{}.tar.gz'.format(
        version),
    packages=setuptools.find_packages(),
    keywords=['gcp', 'google cloud platform', 'framework'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3'
    ],
    include_package_data=True,
    install_requires=[
        'pandas>=1.0.1',
        'setuptools>=36.5.0',
        'google-cloud-storage>=1.26.0',
        'google-cloud-dataproc>=0.6.1',
        'google-auth>=1.11.2',
        'crc32c>=2.0',
        'pyarrow>=0.16.0',
        'google-api-python-client>=1.7.11'
    ],
    python_requires='>=3'
)



"""
Deploy module - deploys code files to bucket into specified dirs.
"""

# pylint: disable=invalid-name, too-many-arguments, too-many-instance-attributes
# pylint: disable=too-few-public-methods, anomalous-backslash-in-string

import re

from glob import glob

from gcpinfra.gcp.conf import GCPConf
from gcpinfra.gcs.comms import GCSComms

class GCSDeployer:
    """
    Deploys codes to a GCS dir, replacing the old ones.

    Does not delete files not listed in the local dir.
    Replaces matched files in the remote dir.

    Files that should be ignore must be informed in 'skip' parameter.
    """

    def __init__(self, bucketname, destinationdir, localdir, skip=None,
                 topleveldir=None, verbose=False):
        """Constructor."""

        self.verbose = verbose
        self.bucketname = bucketname
        self.destinationdir = self.__trailing_backslash(destinationdir)
        self.localdir = self.__trailing_backslash(localdir)
        self.topleveldir = self.__trailing_backslash(topleveldir)
        self.skip = skip

        if self.skip and not isinstance(skip, list):
            self.skip = [self.skip]
        elif not self.skip:
            self.skip = []

        self.__prepare_regex()
        self.files_deploy = self.__find_files()

        self.conf = GCPConf(verbose=verbose)
        self.comms = GCSComms(bucketname, verbose=verbose)

        self.__print('Using top level dir {}'.format(self.topleveldir))

    def __print(self, msg):
        """Prints message accordingly to 'verbose' param."""

        if self.verbose:
            print(msg)

    @staticmethod
    def __trailing_backslash(path):
        """Removes blackslash from variables that contains paths."""

        if not path:
            return path

        if path[-1] != '/':
            return path

        return path[:-1]

    def __prepare_regex(self):
        """Prepares patterns to match."""

        pttrns = []
        for rg in self.skip:
            rg = rg.replace('.', '\.').replace('*', '.*')
            pttrns.append(re.compile(rg))
            self.__print('Skipping pattern {}'.format(rg))

        self.skip = pttrns

    def __find_files(self):
        """Find files to deploy."""

        filepttn = re.compile('{}/(.*\.*)$'.format(self.topleveldir))
        res = []
        files = glob('{}/**/*.*'.format(self.localdir), recursive=True)
        for _file in files:
            _file = _file.replace('./', '')
            remove = False
            for skip_pttrn in self.skip:
                if skip_pttrn.match(_file):
                    remove = True
                    break
            if not remove:
                relativepath = filepttn.match(_file).group(1)
                res.append((_file, relativepath))
                self.__print('{} found'.format(_file))

        return res

    def deploy(self):
        """
        Deploys files to GCS destination.
        """

        self.__print('Starting deploy')
        for _file, relativepath in self.files_deploy:
            path = '{}/{}'.format(self.destinationdir, relativepath)
            self.comms.upload_file(_file, path, replacepolicy='replace')
